"""
    DPN for ImageNet-1K, implemented in TensorFlow.
    Original paper: 'Dual Path Networks,' https://arxiv.org/abs/1707.01629.
"""

__all__ = ['DPN', 'dpn68', 'dpn68b', 'dpn98', 'dpn107', 'dpn131']

import os
import tensorflow as tf
import tensorflow.keras.layers as nn
from .common import MaxPool2d, GlobalAvgPool2d, BatchNorm, Conv2d, conv1x1, DualPathSequential, SimpleSequential,\
    flatten, is_channels_first, get_channel_axis


class GlobalAvgMaxPool2D(nn.Layer):
    """
    Global average+max pooling operation for spatial data.

    Parameters:
    ----------
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 data_format="channels_last",
                 **kwargs):
        super(GlobalAvgMaxPool2D, self).__init__(**kwargs)
        self.axis = get_channel_axis(data_format)

        self.avg_pool = nn.GlobalAvgPool2D(
            data_format=data_format,
            name="avg_pool")
        self.max_pool = nn.GlobalMaxPool2D(
            data_format=data_format,
            name="max_pool")

    def call(self, x, training=None):
        x_avg = self.avg_pool(x)
        x_max = self.max_pool(x)
        x = 0.5 * (x_avg + x_max)
        x = tf.expand_dims(tf.expand_dims(x, axis=self.axis), axis=self.axis)
        return x


def dpn_batch_norm(channels,
                   data_format="channels_last",
                   **kwargs):
    """
    DPN specific Batch normalization layer.

    Parameters:
    ----------
    channels : int
        Number of channels in input data.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    assert (channels is not None)
    return BatchNorm(
        epsilon=0.001,
        data_format=data_format,
        **kwargs)


class PreActivation(nn.Layer):
    """
    DPN specific block, which performs the preactivation like in RreResNet.

    Parameters:
    ----------
    channels : int
        Number of channels.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 channels,
                 data_format="channels_last",
                 **kwargs):
        super(PreActivation, self).__init__(**kwargs)
        self.bn = dpn_batch_norm(
            channels=channels,
            data_format=data_format,
            name="bn")
        self.activ = nn.ReLU()

    def call(self, x, training=None):
        x = self.bn(x, training=training)
        x = self.activ(x)
        return x


class DPNConv(nn.Layer):
    """
    DPN specific convolution block.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    kernel_size : int or tuple/list of 2 int
        Convolution window size.
    strides : int or tuple/list of 2 int
        Strides of the convolution.
    padding : int or tuple/list of 2 int
        Padding value for convolution layer.
    groups : int
        Number of groups.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 strides,
                 padding,
                 groups,
                 data_format="channels_last",
                 **kwargs):
        super(DPNConv, self).__init__(**kwargs)
        self.bn = dpn_batch_norm(
            channels=in_channels,
            data_format=data_format,
            name="bn")
        self.activ = nn.ReLU()
        self.conv = Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            groups=groups,
            use_bias=False,
            data_format=data_format,
            name="conv")

    def call(self, x, training=None):
        x = self.bn(x, training=training)
        x = self.activ(x)
        x = self.conv(x)
        return x


def dpn_conv1x1(in_channels,
                out_channels,
                strides=1,
                data_format="channels_last",
                **kwargs):
    """
    1x1 version of the DPN specific convolution block.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    strides : int or tuple/list of 2 int, default 1
        Strides of the convolution.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    return DPNConv(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=1,
        strides=strides,
        padding=0,
        groups=1,
        data_format=data_format,
        **kwargs)


def dpn_conv3x3(in_channels,
                out_channels,
                strides,
                groups,
                data_format="channels_last",
                **kwargs):
    """
    3x3 version of the DPN specific convolution block.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    strides : int or tuple/list of 2 int
        Strides of the convolution.
    groups : int
        Number of groups.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    return DPNConv(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=3,
        strides=strides,
        padding=1,
        groups=groups,
        data_format=data_format,
        **kwargs)


class DPNUnit(nn.Layer):
    """
    DPN unit.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    mid_channels : int
        Number of intermediate channels.
    bw : int
        Number of residual channels.
    inc : int
        Incrementing step for channels.
    groups : int
        Number of groups in the units.
    has_proj : bool
        Whether to use projection.
    key_strides : int
        Key strides of the convolutions.
    b_case : bool, default False
        Whether to use B-case model.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 mid_channels,
                 bw,
                 inc,
                 groups,
                 has_proj,
                 key_strides,
                 b_case=False,
                 data_format="channels_last",
                 **kwargs):
        super(DPNUnit, self).__init__(**kwargs)
        self.bw = bw
        self.has_proj = has_proj
        self.b_case = b_case
        self.data_format = data_format

        if self.has_proj:
            self.conv_proj = dpn_conv1x1(
                in_channels=in_channels,
                out_channels=bw + 2 * inc,
                strides=key_strides,
                data_format=data_format,
                name="conv_proj")

        self.conv1 = dpn_conv1x1(
            in_channels=in_channels,
            out_channels=mid_channels,
            data_format=data_format,
            name="conv1")
        self.conv2 = dpn_conv3x3(
            in_channels=mid_channels,
            out_channels=mid_channels,
            strides=key_strides,
            groups=groups,
            data_format=data_format,
            name="conv2")

        if b_case:
            self.preactiv = PreActivation(
                channels=mid_channels,
                data_format=data_format,
                name="preactiv")
            self.conv3a = conv1x1(
                in_channels=mid_channels,
                out_channels=bw,
                data_format=data_format,
                name="conv3a")
            self.conv3b = conv1x1(
                in_channels=mid_channels,
                out_channels=inc,
                data_format=data_format,
                name="conv3b")
        else:
            self.conv3 = dpn_conv1x1(
                in_channels=mid_channels,
                out_channels=bw + inc,
                data_format=data_format,
                name="conv3")

    def call(self, x1, x2=None, training=None):
        axis = get_channel_axis(self.data_format)
        x_in = tf.concat([x1, x2], axis=axis) if x2 is not None else x1
        if self.has_proj:
            x_s = self.conv_proj(x_in, training=training)
            channels = (x_s.get_shape().as_list())[axis]
            x_s1, x_s2 = tf.split(x_s, num_or_size_splits=[self.bw, channels - self.bw], axis=axis)
            # x_s1 = F.slice_axis(x_s, axis=1, begin=0, end=self.bw)
            # x_s2 = F.slice_axis(x_s, axis=1, begin=self.bw, end=None)
        else:
            assert (x2 is not None)
            x_s1 = x1
            x_s2 = x2
        x_in = self.conv1(x_in, training=training)
        x_in = self.conv2(x_in, training=training)
        if self.b_case:
            x_in = self.preactiv(x_in, training=training)
            y1 = self.conv3a(x_in, training=training)
            y2 = self.conv3b(x_in, training=training)
        else:
            x_in = self.conv3(x_in, training=training)
            # y1 = F.slice_axis(x_in, axis=1, begin=0, end=self.bw)
            # y2 = F.slice_axis(x_in, axis=1, begin=self.bw, end=None)
            channels = (x_in.get_shape().as_list())[axis]
            y1, y2 = tf.split(x_in, num_or_size_splits=[self.bw, channels - self.bw], axis=axis)
        residual = x_s1 + y1
        dense = tf.concat([x_s2, y2], axis=axis)
        return residual, dense


class DPNInitBlock(nn.Layer):
    """
    DPN specific initial block.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    kernel_size : int or tuple/list of 2 int
        Convolution window size.
    padding : int or tuple/list of 2 int
        Padding value for convolution layer.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 out_channels,
                 kernel_size,
                 padding,
                 data_format="channels_last",
                 **kwargs):
        super(DPNInitBlock, self).__init__(**kwargs)
        self.conv = Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            strides=2,
            padding=padding,
            use_bias=False,
            data_format=data_format,
            name="conv")
        self.bn = dpn_batch_norm(
            channels=out_channels,
            data_format=data_format,
            name="bn")
        self.activ = nn.ReLU()
        self.pool = MaxPool2d(
            pool_size=3,
            strides=2,
            padding=1,
            data_format=data_format,
            name="pool")

    def call(self, x, training=None):
        x = self.conv(x)
        x = self.bn(x, training=training)
        x = self.activ(x)
        x = self.pool(x)
        return x


class DPNFinalBlock(nn.Layer):
    """
    DPN final block, which performs the preactivation with cutting.

    Parameters:
    ----------
    channels : int
        Number of channels.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 channels,
                 data_format="channels_last",
                 **kwargs):
        super(DPNFinalBlock, self).__init__(**kwargs)
        self.data_format = data_format

        self.activ = PreActivation(
            channels=channels,
            data_format=data_format,
            name="activ")

    def call(self, x1, x2, training=None):
        assert (x2 is not None)
        x = tf.concat([x1, x2], axis=get_channel_axis(self.data_format))
        x = self.activ(x)
        return x, None


class DPN(tf.keras.Model):
    """
    DPN model from 'Dual Path Networks,' https://arxiv.org/abs/1707.01629.

    Parameters:
    ----------
    channels : list of list of int
        Number of output channels for each unit.
    init_block_channels : int
        Number of output channels for the initial unit.
    init_block_kernel_size : int or tuple/list of 2 int
        Convolution window size for the initial unit.
    init_block_padding : int or tuple/list of 2 int
        Padding value for convolution layer in the initial unit.
    rs : list f int
        Number of intermediate channels for each unit.
    bws : list f int
        Number of residual channels for each unit.
    incs : list f int
        Incrementing step for channels for each unit.
    groups : int
        Number of groups in the units.
    b_case : bool
        Whether to use B-case model.
    for_training : bool
        Whether to use model for training.
    test_time_pool : bool
        Whether to use the avg-max pooling in the inference mode.
    in_channels : int, default 3
        Number of input channels.
    in_size : tuple of two ints, default (224, 224)
        Spatial size of the expected input image.
    classes : int, default 1000
        Number of classification classes.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 channels,
                 init_block_channels,
                 init_block_kernel_size,
                 init_block_padding,
                 rs,
                 bws,
                 incs,
                 groups,
                 b_case,
                 for_training,
                 test_time_pool,
                 in_channels=3,
                 in_size=(224, 224),
                 classes=1000,
                 data_format="channels_last",
                 **kwargs):
        super(DPN, self).__init__(**kwargs)
        self.in_size = in_size
        self.classes = classes
        self.data_format = data_format

        self.features = DualPathSequential(
            return_two=False,
            first_ordinals=1,
            last_ordinals=0,
            name="features")
        self.features.children.append(DPNInitBlock(
            in_channels=in_channels,
            out_channels=init_block_channels,
            kernel_size=init_block_kernel_size,
            padding=init_block_padding,
            data_format=data_format,
            name="init_block"))
        in_channels = init_block_channels
        for i, channels_per_stage in enumerate(channels):
            stage = DualPathSequential(name="stage{}".format(i + 1))
            r = rs[i]
            bw = bws[i]
            inc = incs[i]
            for j, out_channels in enumerate(channels_per_stage):
                has_proj = (j == 0)
                key_strides = 2 if (j == 0) and (i != 0) else 1
                stage.children.append(DPNUnit(
                    in_channels=in_channels,
                    mid_channels=r,
                    bw=bw,
                    inc=inc,
                    groups=groups,
                    has_proj=has_proj,
                    key_strides=key_strides,
                    b_case=b_case,
                    data_format=data_format,
                    name="unit{}".format(j + 1)))
                in_channels = out_channels
            self.features.children.append(stage)
        self.features.children.append(DPNFinalBlock(
            channels=in_channels,
            data_format=data_format,
            name="final_block"))

        self.output1 = SimpleSequential(name="output1")
        if for_training or not test_time_pool:
            self.output1.add(GlobalAvgPool2d(
                data_format=data_format,
                name="final_pool"))
            self.output1.add(conv1x1(
                in_channels=in_channels,
                out_channels=classes,
                use_bias=True,
                data_format=data_format,
                name="classifier"))
        else:
            self.output1.add(nn.AveragePooling2D(
                pool_size=7,
                strides=1,
                data_format=data_format,
                name="avg_pool"))
            self.output1.add(conv1x1(
                in_channels=in_channels,
                out_channels=classes,
                use_bias=True,
                data_format=data_format,
                name="classifier"))
            self.output1.add(GlobalAvgMaxPool2D(
                data_format=data_format,
                name="avgmax_pool"))

    def call(self, x, training=None):
        x = self.features(x, training=training)
        x = self.output1(x)
        x = flatten(x, self.data_format)
        return x


def get_dpn(num_layers,
            b_case=False,
            for_training=False,
            model_name=None,
            pretrained=False,
            root=os.path.join("~", ".tensorflow", "models"),
            **kwargs):
    """
    Create DPN model with specific parameters.

    Parameters:
    ----------
    num_layers : int
        Number of layers.
    b_case : bool, default False
        Whether to use B-case model.
    for_training : bool
        Whether to use model for training.
    model_name : str or None, default None
        Model name for loading pretrained model.
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    root : str, default '~/.tensorflow/models'
        Location for keeping the model parameters.
    """

    if num_layers == 68:
        init_block_channels = 10
        init_block_kernel_size = 3
        init_block_padding = 1
        bw_factor = 1
        k_r = 128
        groups = 32
        k_sec = (3, 4, 12, 3)
        incs = (16, 32, 32, 64)
        test_time_pool = True
    elif num_layers == 98:
        init_block_channels = 96
        init_block_kernel_size = 7
        init_block_padding = 3
        bw_factor = 4
        k_r = 160
        groups = 40
        k_sec = (3, 6, 20, 3)
        incs = (16, 32, 32, 128)
        test_time_pool = True
    elif num_layers == 107:
        init_block_channels = 128
        init_block_kernel_size = 7
        init_block_padding = 3
        bw_factor = 4
        k_r = 200
        groups = 50
        k_sec = (4, 8, 20, 3)
        incs = (20, 64, 64, 128)
        test_time_pool = True
    elif num_layers == 131:
        init_block_channels = 128
        init_block_kernel_size = 7
        init_block_padding = 3
        bw_factor = 4
        k_r = 160
        groups = 40
        k_sec = (4, 8, 28, 3)
        incs = (16, 32, 32, 128)
        test_time_pool = True
    else:
        raise ValueError("Unsupported DPN version with number of layers {}".format(num_layers))

    channels = [[0] * li for li in k_sec]
    rs = [0 * li for li in k_sec]
    bws = [0 * li for li in k_sec]
    for i in range(len(k_sec)):
        rs[i] = (2 ** i) * k_r
        bws[i] = (2 ** i) * 64 * bw_factor
        inc = incs[i]
        channels[i][0] = bws[i] + 3 * inc
        for j in range(1, k_sec[i]):
            channels[i][j] = channels[i][j - 1] + inc

    net = DPN(
        channels=channels,
        init_block_channels=init_block_channels,
        init_block_kernel_size=init_block_kernel_size,
        init_block_padding=init_block_padding,
        rs=rs,
        bws=bws,
        incs=incs,
        groups=groups,
        b_case=b_case,
        for_training=for_training,
        test_time_pool=test_time_pool,
        **kwargs)

    if pretrained:
        if (model_name is None) or (not model_name):
            raise ValueError("Parameter `model_name` should be properly initialized for loading pretrained model.")
        from .model_store import get_model_file
        in_channels = kwargs["in_channels"] if ("in_channels" in kwargs) else 3
        input_shape = (1,) + (in_channels,) + net.in_size if net.data_format == "channels_first" else\
            (1,) + net.in_size + (in_channels,)
        net.build(input_shape=input_shape)
        net.load_weights(
            filepath=get_model_file(
                model_name=model_name,
                local_model_store_dir_path=root))

    return net


def dpn68(**kwargs):
    """
    DPN-68 model from 'Dual Path Networks,' https://arxiv.org/abs/1707.01629.

    Parameters:
    ----------
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    root : str, default '~/.tensorflow/models'
        Location for keeping the model parameters.
    """
    return get_dpn(num_layers=68, b_case=False, model_name="dpn68", **kwargs)


def dpn68b(**kwargs):
    """
    DPN-68b model from 'Dual Path Networks,' https://arxiv.org/abs/1707.01629.

    Parameters:
    ----------
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    root : str, default '~/.tensorflow/models'
        Location for keeping the model parameters.
    """
    return get_dpn(num_layers=68, b_case=True, model_name="dpn68b", **kwargs)


def dpn98(**kwargs):
    """
    DPN-98 model from 'Dual Path Networks,' https://arxiv.org/abs/1707.01629.

    Parameters:
    ----------
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    root : str, default '~/.tensorflow/models'
        Location for keeping the model parameters.
    """
    return get_dpn(num_layers=98, b_case=False, model_name="dpn98", **kwargs)


def dpn107(**kwargs):
    """
    DPN-107 model from 'Dual Path Networks,' https://arxiv.org/abs/1707.01629.

    Parameters:
    ----------
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    root : str, default '~/.tensorflow/models'
        Location for keeping the model parameters.
    """
    return get_dpn(num_layers=107, b_case=False, model_name="dpn107", **kwargs)


def dpn131(**kwargs):
    """
    DPN-131 model from 'Dual Path Networks,' https://arxiv.org/abs/1707.01629.

    Parameters:
    ----------
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    root : str, default '~/.tensorflow/models'
        Location for keeping the model parameters.
    """
    return get_dpn(num_layers=131, b_case=False, model_name="dpn131", **kwargs)


def _test():
    import numpy as np
    import tensorflow.keras.backend as K

    data_format = "channels_last"
    pretrained = False

    models = [
        dpn68,
        dpn68b,
        dpn98,
        dpn107,
        dpn131,
    ]

    for model in models:

        net = model(pretrained=pretrained, data_format=data_format)

        batch = 14
        x = tf.random.normal((batch, 3, 224, 224) if is_channels_first(data_format) else (batch, 224, 224, 3))
        y = net(x)
        assert (tuple(y.shape.as_list()) == (batch, 1000))

        weight_count = sum([np.prod(K.get_value(w).shape) for w in net.trainable_weights])
        print("m={}, {}".format(model.__name__, weight_count))
        assert (model != dpn68 or weight_count == 12611602)
        assert (model != dpn68b or weight_count == 12611602)
        assert (model != dpn98 or weight_count == 61570728)
        assert (model != dpn107 or weight_count == 86917800)
        assert (model != dpn131 or weight_count == 79254504)


if __name__ == "__main__":
    _test()



"""
    CGNet for image segmentation, implemented in TensorFlow.
    Original paper: 'CGNet: A Light-weight Context Guided Network for Semantic Segmentation,'
    https://arxiv.org/abs/1811.08201.
"""

__all__ = ['CGNet', 'cgnet_cityscapes']

import os
import tensorflow as tf
import tensorflow.keras.layers as nn
from .common import NormActivation, conv1x1, conv1x1_block, conv3x3_block, depthwise_conv3x3, SEBlock, Concurrent,\
    DualPathSequential, InterpolationBlock, SimpleSequential, is_channels_first, get_im_size, PReLU2, AvgPool2d,\
    get_channel_axis


class CGBlock(nn.Layer):
    """
    CGNet block.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    dilation : int
        Dilation value.
    se_reduction : int
        SE-block reduction value.
    down : bool
        Whether to downsample.
    bn_eps : float
        Small float added to variance in Batch norm.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 out_channels,
                 dilation,
                 se_reduction,
                 down,
                 bn_eps,
                 data_format="channels_last",
                 **kwargs):
        super(CGBlock, self).__init__(**kwargs)
        self.down = down
        if self.down:
            mid1_channels = out_channels
            mid2_channels = 2 * out_channels
        else:
            mid1_channels = out_channels // 2
            mid2_channels = out_channels

        if self.down:
            self.conv1 = conv3x3_block(
                in_channels=in_channels,
                out_channels=out_channels,
                strides=2,
                bn_eps=bn_eps,
                activation=(lambda: PReLU2(out_channels, data_format=data_format, name="activ")),
                data_format=data_format,
                name="conv1")
        else:
            self.conv1 = conv1x1_block(
                in_channels=in_channels,
                out_channels=mid1_channels,
                bn_eps=bn_eps,
                activation=(lambda: PReLU2(mid1_channels, data_format=data_format, name="activ")),
                data_format=data_format,
                name="conv1")

        self.branches = Concurrent(
            data_format=data_format,
            name="branches")
        self.branches.add(depthwise_conv3x3(
            channels=mid1_channels,
            data_format=data_format,
            name="branches1"))
        self.branches.add(depthwise_conv3x3(
            channels=mid1_channels,
            padding=dilation,
            dilation=dilation,
            data_format=data_format,
            name="branches2"))

        self.norm_activ = NormActivation(
            in_channels=mid2_channels,
            bn_eps=bn_eps,
            activation=(lambda: PReLU2(mid2_channels, data_format=data_format, name="activ")),
            data_format=data_format,
            name="norm_activ")

        if self.down:
            self.conv2 = conv1x1(
                in_channels=mid2_channels,
                out_channels=out_channels,
                data_format=data_format,
                name="conv2")

        self.se = SEBlock(
            channels=out_channels,
            reduction=se_reduction,
            use_conv=False,
            data_format=data_format,
            name="se")

    def call(self, x, training=None):
        if not self.down:
            identity = x
        x = self.conv1(x, training=training)
        x = self.branches(x, training=training)
        x = self.norm_activ(x, training=training)
        if self.down:
            x = self.conv2(x, training=training)
        x = self.se(x, training=training)
        if not self.down:
            x += identity
        return x


class CGUnit(nn.Layer):
    """
    CGNet unit.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    layers : int
        Number of layers.
    dilation : int
        Dilation value.
    se_reduction : int
        SE-block reduction value.
    bn_eps : float
        Small float added to variance in Batch norm.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 out_channels,
                 layers,
                 dilation,
                 se_reduction,
                 bn_eps,
                 data_format="channels_last",
                 **kwargs):
        super(CGUnit, self).__init__(**kwargs)
        self.axis = get_channel_axis(data_format)
        mid_channels = out_channels // 2

        self.down = CGBlock(
            in_channels=in_channels,
            out_channels=mid_channels,
            dilation=dilation,
            se_reduction=se_reduction,
            down=True,
            bn_eps=bn_eps,
            data_format=data_format,
            name="down")
        self.blocks = SimpleSequential(name="blocks")
        for i in range(layers - 1):
            self.blocks.add(CGBlock(
                in_channels=mid_channels,
                out_channels=mid_channels,
                dilation=dilation,
                se_reduction=se_reduction,
                down=False,
                bn_eps=bn_eps,
                data_format=data_format,
                name="block{}".format(i + 1)))

    def call(self, x, training=None):
        x = self.down(x, training=training)
        y = self.blocks(x, training=training)
        x = tf.concat([y, x], axis=self.axis)  # NB: This differs from the original implementation.
        return x


class CGStage(nn.Layer):
    """
    CGNet stage.

    Parameters:
    ----------
    x_channels : int
        Number of input/output channels for x.
    y_in_channels : int
        Number of input channels for y.
    y_out_channels : int
        Number of output channels for y.
    layers : int
        Number of layers in the unit.
    dilation : int
        Dilation for blocks.
    se_reduction : int
        SE-block reduction value for blocks.
    bn_eps : float
        Small float added to variance in Batch norm.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 x_channels,
                 y_in_channels,
                 y_out_channels,
                 layers,
                 dilation,
                 se_reduction,
                 bn_eps,
                 data_format="channels_last",
                 **kwargs):
        super(CGStage, self).__init__(**kwargs)
        self.axis = get_channel_axis(data_format)
        self.use_x = (x_channels > 0)
        self.use_unit = (layers > 0)

        if self.use_x:
            self.x_down = AvgPool2d(
                pool_size=3,
                strides=2,
                padding=1,
                data_format=data_format,
                name="x_down")

        if self.use_unit:
            self.unit = CGUnit(
                in_channels=y_in_channels,
                out_channels=(y_out_channels - x_channels),
                layers=layers,
                dilation=dilation,
                se_reduction=se_reduction,
                bn_eps=bn_eps,
                data_format=data_format,
                name="unit")

        self.norm_activ = NormActivation(
            in_channels=y_out_channels,
            bn_eps=bn_eps,
            activation=(lambda: PReLU2(y_out_channels, data_format=data_format, name="activ")),
            data_format=data_format,
            name="norm_activ")

    def call(self, y, x=None, training=None):
        if self.use_unit:
            y = self.unit(y, training=training)
        if self.use_x:
            x = self.x_down(x)
            y = tf.concat([y, x], axis=self.axis)
        y = self.norm_activ(y, training=training)
        return y, x


class CGInitBlock(nn.Layer):
    """
    CGNet specific initial block.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    bn_eps : float
        Small float added to variance in Batch norm.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 out_channels,
                 bn_eps,
                 data_format="channels_last",
                 **kwargs):
        super(CGInitBlock, self).__init__(**kwargs)
        self.conv1 = conv3x3_block(
            in_channels=in_channels,
            out_channels=out_channels,
            strides=2,
            bn_eps=bn_eps,
            activation=(lambda: PReLU2(out_channels, data_format=data_format, name="activ")),
            data_format=data_format,
            name="conv1")
        self.conv2 = conv3x3_block(
            in_channels=out_channels,
            out_channels=out_channels,
            bn_eps=bn_eps,
            activation=(lambda: PReLU2(out_channels, data_format=data_format, name="activ")),
            data_format=data_format,
            name="conv2")
        self.conv3 = conv3x3_block(
            in_channels=out_channels,
            out_channels=out_channels,
            bn_eps=bn_eps,
            activation=(lambda: PReLU2(out_channels, data_format=data_format, name="activ")),
            data_format=data_format,
            name="conv3")

    def call(self, x, training=None):
        x = self.conv1(x, training=training)
        x = self.conv2(x, training=training)
        x = self.conv3(x, training=training)
        return x


class CGNet(tf.keras.Model):
    """
    CGNet model from 'CGNet: A Light-weight Context Guided Network for Semantic Segmentation,'
    https://arxiv.org/abs/1811.08201.

    Parameters:
    ----------
    layers : list of int
        Number of layers for each unit.
    channels : list of int
        Number of output channels for each unit (for y-branch).
    init_block_channels : int
        Number of output channels for the initial unit.
    dilations : list of int
        Dilations for each unit.
    se_reductions : list of int
        SE-block reduction value for each unit.
    cut_x : list of int
        Whether to concatenate with x-branch for each unit.
    bn_eps : float, default 1e-5
        Small float added to variance in Batch norm.
    aux : bool, default False
        Whether to output an auxiliary result.
    fixed_size : bool, default False
        Whether to expect fixed spatial size of input image.
    in_channels : int, default 3
        Number of input channels.
    in_size : tuple of two ints, default (1024, 2048)
        Spatial size of the expected input image.
    classes : int, default 19
        Number of segmentation classes.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 layers,
                 channels,
                 init_block_channels,
                 dilations,
                 se_reductions,
                 cut_x,
                 bn_eps=1e-5,
                 aux=False,
                 fixed_size=False,
                 in_channels=3,
                 in_size=(1024, 2048),
                 classes=19,
                 data_format="channels_last",
                 **kwargs):
        super(CGNet, self).__init__(**kwargs)
        assert (aux is not None)
        assert (fixed_size is not None)
        assert ((in_size[0] % 8 == 0) and (in_size[1] % 8 == 0))
        self.in_size = in_size
        self.classes = classes
        self.fixed_size = fixed_size
        self.data_format = data_format

        self.features = DualPathSequential(
            return_two=False,
            first_ordinals=1,
            last_ordinals=0,
            name="features")
        self.features.add(CGInitBlock(
            in_channels=in_channels,
            out_channels=init_block_channels,
            bn_eps=bn_eps,
            data_format=data_format,
            name="init_block"))
        y_in_channels = init_block_channels

        for i, (layers_i, y_out_channels) in enumerate(zip(layers, channels)):
            self.features.add(CGStage(
                x_channels=in_channels if cut_x[i] == 1 else 0,
                y_in_channels=y_in_channels,
                y_out_channels=y_out_channels,
                layers=layers_i,
                dilation=dilations[i],
                se_reduction=se_reductions[i],
                bn_eps=bn_eps,
                data_format=data_format,
                name="stage{}".format(i + 1)))
            y_in_channels = y_out_channels

        self.classifier = conv1x1(
            in_channels=y_in_channels,
            out_channels=classes,
            data_format=data_format,
            name="classifier")

        self.up = InterpolationBlock(
            scale_factor=8,
            data_format=data_format,
            name="up")

    def call(self, x, training=None):
        in_size = self.in_size if self.fixed_size else get_im_size(x, data_format=self.data_format)
        y = self.features(x, x, training=training)
        y = self.classifier(y)
        y = self.up(y, size=in_size)
        return y


def get_cgnet(model_name=None,
              pretrained=False,
              root=os.path.join("~", ".tensorflow", "models"),
              **kwargs):
    """
    Create CGNet model with specific parameters.

    Parameters:
    ----------
    model_name : str or None, default None
        Model name for loading pretrained model.
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    root : str, default '~/.tensorflow/models'
        Location for keeping the model parameters.
    """
    init_block_channels = 32
    layers = [0, 3, 21]
    channels = [35, 131, 256]
    dilations = [0, 2, 4]
    se_reductions = [0, 8, 16]
    cut_x = [1, 1, 0]
    bn_eps = 1e-3

    net = CGNet(
        layers=layers,
        channels=channels,
        init_block_channels=init_block_channels,
        dilations=dilations,
        se_reductions=se_reductions,
        cut_x=cut_x,
        bn_eps=bn_eps,
        **kwargs)

    if pretrained:
        if (model_name is None) or (not model_name):
            raise ValueError("Parameter `model_name` should be properly initialized for loading pretrained model.")
        from .model_store import get_model_file
        in_channels = kwargs["in_channels"] if ("in_channels" in kwargs) else 3
        input_shape = (1,) + (in_channels,) + net.in_size if net.data_format == "channels_first" else\
            (1,) + net.in_size + (in_channels,)
        net.build(input_shape=input_shape)
        net.load_weights(
            filepath=get_model_file(
                model_name=model_name,
                local_model_store_dir_path=root),
            by_name=True,
            skip_mismatch=True)

    return net


def cgnet_cityscapes(classes=19, **kwargs):
    """
    CGNet model for Cityscapes from 'CGNet: A Light-weight Context Guided Network for Semantic Segmentation,'
    https://arxiv.org/abs/1811.08201.

    Parameters:
    ----------
    classes : int, default 19
        Number of segmentation classes.
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    root : str, default '~/.tensorflow/models'
        Location for keeping the model parameters.
    """
    return get_cgnet(classes=classes, model_name="cgnet_cityscapes", **kwargs)


def _test():
    import numpy as np
    import tensorflow.keras.backend as K

    data_format = "channels_last"
    # data_format = "channels_first"
    pretrained = False
    in_size = (1024, 2048)
    classes = 19

    models = [
        cgnet_cityscapes,
    ]

    for model in models:

        net = model(pretrained=pretrained, in_size=in_size, data_format=data_format)

        batch = 4
        x = tf.random.normal((batch, 3, in_size[0], in_size[1]) if is_channels_first(data_format) else
                             (batch, in_size[0], in_size[1], 3))
        y = net(x)
        assert (tuple(y.shape.as_list()) == (batch, classes, in_size[0], in_size[1]) if is_channels_first(data_format)
                else tuple(y.shape.as_list()) == (batch, in_size[0], in_size[1], classes))

        weight_count = sum([np.prod(K.get_value(w).shape) for w in net.trainable_weights])
        print("m={}, {}".format(model.__name__, weight_count))
        assert (model != cgnet_cityscapes or weight_count == 496306)


if __name__ == "__main__":
    _test()



"""
    Lightweight OpenPose 2D/3D for CMU Panoptic, implemented in TensorFlow.
    Original paper: 'Real-time 2D Multi-Person Pose Estimation on CPU: Lightweight OpenPose,'
    https://arxiv.org/abs/1811.12004.
"""

__all__ = ['LwOpenPose', 'lwopenpose2d_mobilenet_cmupan_coco', 'lwopenpose3d_mobilenet_cmupan_coco',
           'LwopDecoderFinalBlock']

import os
import tensorflow as tf
import tensorflow.keras.layers as nn
from .common import conv1x1, conv1x1_block, conv3x3_block, dwsconv3x3_block, SimpleSequential, is_channels_first,\
    get_channel_axis


class LwopResBottleneck(nn.Layer):
    """
    Bottleneck block for residual path in the residual unit.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    strides : int or tuple/list of 2 int
        Strides of the convolution.
    use_bias : bool, default True
        Whether the layer uses a bias vector.
    bottleneck_factor : int, default 2
        Bottleneck factor.
    squeeze_out : bool, default False
        Whether to squeeze the output channels.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 out_channels,
                 strides,
                 use_bias=True,
                 bottleneck_factor=2,
                 squeeze_out=False,
                 data_format="channels_last",
                 **kwargs):
        super(LwopResBottleneck, self).__init__(**kwargs)
        mid_channels = out_channels // bottleneck_factor if squeeze_out else in_channels // bottleneck_factor

        self.conv1 = conv1x1_block(
            in_channels=in_channels,
            out_channels=mid_channels,
            use_bias=use_bias,
            data_format=data_format,
            name="conv1")
        self.conv2 = conv3x3_block(
            in_channels=mid_channels,
            out_channels=mid_channels,
            strides=strides,
            use_bias=use_bias,
            data_format=data_format,
            name="conv2")
        self.conv3 = conv1x1_block(
            in_channels=mid_channels,
            out_channels=out_channels,
            use_bias=use_bias,
            activation=None,
            data_format=data_format,
            name="conv3")

    def call(self, x, training=None):
        x = self.conv1(x, training=training)
        x = self.conv2(x, training=training)
        x = self.conv3(x, training=training)
        return x


class LwopResUnit(nn.Layer):
    """
    ResNet-like residual unit with residual connection.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    strides : int or tuple/list of 2 int, default 1
        Strides of the convolution.
    use_bias : bool, default True
        Whether the layer uses a bias vector.
    bottleneck_factor : int, default 2
        Bottleneck factor.
    squeeze_out : bool, default False
        Whether to squeeze the output channels.
    activate : bool, default False
        Whether to activate the sum.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 out_channels,
                 strides=1,
                 use_bias=True,
                 bottleneck_factor=2,
                 squeeze_out=False,
                 activate=False,
                 data_format="channels_last",
                 **kwargs):
        super(LwopResUnit, self).__init__(**kwargs)
        self.activate = activate
        self.resize_identity = (in_channels != out_channels) or (strides != 1)

        self.body = LwopResBottleneck(
            in_channels=in_channels,
            out_channels=out_channels,
            strides=strides,
            use_bias=use_bias,
            bottleneck_factor=bottleneck_factor,
            squeeze_out=squeeze_out,
            data_format=data_format,
            name="body")
        if self.resize_identity:
            self.identity_conv = conv1x1_block(
                in_channels=in_channels,
                out_channels=out_channels,
                strides=strides,
                use_bias=use_bias,
                activation=None,
                data_format=data_format,
                name="identity_conv")
        if self.activate:
            self.activ = nn.ReLU()

    def call(self, x, training=None):
        if self.resize_identity:
            identity = self.identity_conv(x, training=training)
        else:
            identity = x
        x = self.body(x, training=training)
        x = x + identity
        if self.activate:
            x = self.activ(x)
        return x


class LwopEncoderFinalBlock(nn.Layer):
    """
    Lightweight OpenPose 2D/3D specific encoder final block.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 out_channels,
                 data_format="channels_last",
                 **kwargs):
        super(LwopEncoderFinalBlock, self).__init__(**kwargs)
        self.pre_conv = conv1x1_block(
            in_channels=in_channels,
            out_channels=out_channels,
            use_bias=True,
            use_bn=False,
            data_format=data_format,
            name="pre_conv")
        self.body = SimpleSequential(name="body")
        for i in range(3):
            self.body.add(dwsconv3x3_block(
                in_channels=out_channels,
                out_channels=out_channels,
                dw_use_bn=False,
                pw_use_bn=False,
                dw_activation=(lambda: nn.ELU()),
                pw_activation=(lambda: nn.ELU()),
                data_format=data_format,
                name="block{}".format(i + 1)))
        self.post_conv = conv3x3_block(
            in_channels=out_channels,
            out_channels=out_channels,
            use_bias=True,
            use_bn=False,
            data_format=data_format,
            name="post_conv")

    def call(self, x, training=None):
        x = self.pre_conv(x, training=training)
        x = x + self.body(x, training=training)
        x = self.post_conv(x, training=training)
        return x


class LwopRefinementBlock(nn.Layer):
    """
    Lightweight OpenPose 2D/3D specific refinement block for decoder units.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 out_channels,
                 data_format="channels_last",
                 **kwargs):
        super(LwopRefinementBlock, self).__init__(**kwargs)
        self.pre_conv = conv1x1_block(
            in_channels=in_channels,
            out_channels=out_channels,
            use_bias=True,
            use_bn=False,
            data_format=data_format,
            name="pre_conv")
        self.body = SimpleSequential(name="body")
        self.body.add(conv3x3_block(
            in_channels=out_channels,
            out_channels=out_channels,
            use_bias=True,
            data_format=data_format,
            name="block1"))
        self.body.add(conv3x3_block(
            in_channels=out_channels,
            out_channels=out_channels,
            padding=2,
            dilation=2,
            use_bias=True,
            data_format=data_format,
            name="block2"))

    def call(self, x, training=None):
        x = self.pre_conv(x, training=training)
        # print("--> x.shape={}".format(x.shape))
        y = self.body(x, training=training)
        # print("==> x.shape={}".format(x.shape))
        x = x + y
        return x


class LwopDecoderBend(nn.Layer):
    """
    Lightweight OpenPose 2D/3D specific decoder bend block.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    mid_channels : int
        Number of middle channels.
    out_channels : int
        Number of output channels.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 mid_channels,
                 out_channels,
                 data_format="channels_last",
                 **kwargs):
        super(LwopDecoderBend, self).__init__(**kwargs)
        self.conv1 = conv1x1_block(
            in_channels=in_channels,
            out_channels=mid_channels,
            use_bias=True,
            use_bn=False,
            data_format=data_format,
            name="conv1")
        self.conv2 = conv1x1(
            in_channels=mid_channels,
            out_channels=out_channels,
            use_bias=True,
            data_format=data_format,
            name="conv2")

    def call(self, x, training=None):
        x = self.conv1(x, training=training)
        x = self.conv2(x, training=training)
        return x


class LwopDecoderInitBlock(nn.Layer):
    """
    Lightweight OpenPose 2D/3D specific decoder init block.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    keypoints : int
        Number of keypoints.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 keypoints,
                 data_format="channels_last",
                 **kwargs):
        super(LwopDecoderInitBlock, self).__init__(**kwargs)
        self.data_format = data_format
        num_heatmap = keypoints
        num_paf = 2 * keypoints
        bend_mid_channels = 512

        self.body = SimpleSequential(name="body")
        for i in range(3):
            self.body.add(conv3x3_block(
                in_channels=in_channels,
                out_channels=in_channels,
                use_bias=True,
                use_bn=False,
                data_format=data_format,
                name="block{}".format(i + 1)))
        self.heatmap_bend = LwopDecoderBend(
            in_channels=in_channels,
            mid_channels=bend_mid_channels,
            out_channels=num_heatmap,
            data_format=data_format,
            name="heatmap_bend")
        self.paf_bend = LwopDecoderBend(
            in_channels=in_channels,
            mid_channels=bend_mid_channels,
            out_channels=num_paf,
            data_format=data_format,
            name="paf_bend")

    def call(self, x, training=None):
        y = self.body(x, training=training)
        heatmap = self.heatmap_bend(y, training=training)
        paf = self.paf_bend(y, training=training)
        y = tf.concat((x, heatmap, paf), axis=get_channel_axis(self.data_format))
        return y


class LwopDecoderUnit(nn.Layer):
    """
    Lightweight OpenPose 2D/3D specific decoder init.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    keypoints : int
        Number of keypoints.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 keypoints,
                 data_format="channels_last",
                 **kwargs):
        super(LwopDecoderUnit, self).__init__(**kwargs)
        self.data_format = data_format
        num_heatmap = keypoints
        num_paf = 2 * keypoints
        self.features_channels = in_channels - num_heatmap - num_paf

        self.body = SimpleSequential(name="body")
        for i in range(5):
            self.body.add(LwopRefinementBlock(
                in_channels=in_channels,
                out_channels=self.features_channels,
                data_format=data_format,
                name="block{}".format(i + 1)))
            in_channels = self.features_channels
        self.heatmap_bend = LwopDecoderBend(
            in_channels=self.features_channels,
            mid_channels=self.features_channels,
            out_channels=num_heatmap,
            data_format=data_format,
            name="heatmap_bend")
        self.paf_bend = LwopDecoderBend(
            in_channels=self.features_channels,
            mid_channels=self.features_channels,
            out_channels=num_paf,
            data_format=data_format,
            name="paf_bend")

    def call(self, x, training=None):
        if is_channels_first(self.data_format):
            features = x[:, :self.features_channels]
        else:
            features = x[:, :, :, :self.features_channels]
        y = self.body(x, training=training)
        heatmap = self.heatmap_bend(y, training=training)
        paf = self.paf_bend(y, training=training)
        y = tf.concat((features, heatmap, paf), axis=get_channel_axis(self.data_format))
        return y


class LwopDecoderFeaturesBend(nn.Layer):
    """
    Lightweight OpenPose 2D/3D specific decoder 3D features bend.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    mid_channels : int
        Number of middle channels.
    out_channels : int
        Number of output channels.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 mid_channels,
                 out_channels,
                 data_format="channels_last",
                 **kwargs):
        super(LwopDecoderFeaturesBend, self).__init__(**kwargs)
        self.body = SimpleSequential(name="body")
        for i in range(2):
            self.body.add(LwopRefinementBlock(
                in_channels=in_channels,
                out_channels=mid_channels,
                data_format=data_format,
                name="block{}".format(i + 1)))
            in_channels = mid_channels
        self.features_bend = LwopDecoderBend(
            in_channels=mid_channels,
            mid_channels=mid_channels,
            out_channels=out_channels,
            data_format=data_format,
            name="features_bend")

    def call(self, x, training=None):
        x = self.body(x, training=training)
        x = self.features_bend(x, training=training)
        return x


class LwopDecoderFinalBlock(nn.Layer):
    """
    Lightweight OpenPose 2D/3D specific decoder final block for calcualation 3D poses.

    Parameters:
    ----------
    in_channels : int
        Number of input channels.
    keypoints : int
        Number of keypoints.
    bottleneck_factor : int
        Bottleneck factor.
    calc_3d_features : bool
        Whether to calculate 3D features.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 in_channels,
                 keypoints,
                 bottleneck_factor,
                 calc_3d_features,
                 data_format="channels_last",
                 **kwargs):
        super(LwopDecoderFinalBlock, self).__init__(**kwargs)
        self.data_format = data_format
        self.num_heatmap_paf = 3 * keypoints
        self.calc_3d_features = calc_3d_features
        features_out_channels = self.num_heatmap_paf
        features_in_channels = in_channels - features_out_channels

        if self.calc_3d_features:
            self.body = SimpleSequential(name="body")
            for i in range(5):
                self.body.add(LwopResUnit(
                    in_channels=in_channels,
                    out_channels=features_in_channels,
                    bottleneck_factor=bottleneck_factor,
                    data_format=data_format,
                    name="block{}".format(i + 1)))
                in_channels = features_in_channels
            self.features_bend = LwopDecoderFeaturesBend(
                in_channels=features_in_channels,
                mid_channels=features_in_channels,
                out_channels=features_out_channels,
                data_format=data_format,
                name="features_bend")

    def call(self, x, training=None):
        if is_channels_first(self.data_format):
            heatmap_paf_2d = x[:, -self.num_heatmap_paf:]
        else:
            heatmap_paf_2d = x[:, :, :, -self.num_heatmap_paf:]
        if not self.calc_3d_features:
            return heatmap_paf_2d
        x = self.body(x, training=training)
        x = self.features_bend(x, training=training)
        y = tf.concat((heatmap_paf_2d, x), axis=get_channel_axis(self.data_format))
        return y


class LwOpenPose(tf.keras.Model):
    """
    Lightweight OpenPose 2D/3D model from 'Real-time 2D Multi-Person Pose Estimation on CPU: Lightweight OpenPose,'
    https://arxiv.org/abs/1811.12004.

    Parameters:
    ----------
    encoder_channels : list of list of int
        Number of output channels for each encoder unit.
    encoder_paddings : list of list of int
        Padding/dilation value for each encoder unit.
    encoder_init_block_channels : int
        Number of output channels for the encoder initial unit.
    encoder_final_block_channels : int
        Number of output channels for the encoder final unit.
    refinement_units : int
        Number of refinement blocks in the decoder.
    calc_3d_features : bool
        Whether to calculate 3D features.
    return_heatmap : bool, default True
        Whether to return only heatmap.
    in_channels : int, default 3
        Number of input channels.
    in_size : tuple of two ints, default (256, 192)
        Spatial size of the expected input image.
    keypoints : int, default 19
        Number of keypoints.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    """
    def __init__(self,
                 encoder_channels,
                 encoder_paddings,
                 encoder_init_block_channels,
                 encoder_final_block_channels,
                 refinement_units,
                 calc_3d_features,
                 return_heatmap=True,
                 in_channels=3,
                 in_size=(368, 368),
                 keypoints=19,
                 data_format="channels_last",
                 **kwargs):
        super(LwOpenPose, self).__init__(**kwargs)
        assert (in_channels == 3)
        self.in_size = in_size
        self.keypoints = keypoints
        self.data_format = data_format
        self.return_heatmap = return_heatmap
        self.calc_3d_features = calc_3d_features
        num_heatmap_paf = 3 * keypoints

        self.encoder = SimpleSequential(name="encoder")
        backbone = SimpleSequential(name="backbone")
        backbone.add(conv3x3_block(
            in_channels=in_channels,
            out_channels=encoder_init_block_channels,
            strides=2,
            data_format=data_format,
            name="init_block"))
        in_channels = encoder_init_block_channels
        for i, channels_per_stage in enumerate(encoder_channels):
            stage = SimpleSequential(name="stage{}".format(i + 1))
            for j, out_channels in enumerate(channels_per_stage):
                strides = 2 if (j == 0) and (i != 0) else 1
                padding = encoder_paddings[i][j]
                stage.add(dwsconv3x3_block(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    strides=strides,
                    padding=padding,
                    dilation=padding,
                    data_format=data_format,
                    name="unit{}".format(j + 1)))
                in_channels = out_channels
            backbone.add(stage)
        self.encoder.add(backbone)
        self.encoder.add(LwopEncoderFinalBlock(
            in_channels=in_channels,
            out_channels=encoder_final_block_channels,
            data_format=data_format,
            name="final_block"))
        in_channels = encoder_final_block_channels

        self.decoder = SimpleSequential(name="decoder")
        self.decoder.add(LwopDecoderInitBlock(
            in_channels=in_channels,
            keypoints=keypoints,
            data_format=data_format,
            name="init_block"))
        in_channels = encoder_final_block_channels + num_heatmap_paf
        for i in range(refinement_units):
            self.decoder.add(LwopDecoderUnit(
                in_channels=in_channels,
                keypoints=keypoints,
                data_format=data_format,
                name="unit{}".format(i + 1)))
        self.decoder.add(LwopDecoderFinalBlock(
            in_channels=in_channels,
            keypoints=keypoints,
            bottleneck_factor=2,
            calc_3d_features=calc_3d_features,
            data_format=data_format,
            name="final_block"))

    def call(self, x, training=None):
        # print("**> x.shape={}".format(x.shape))
        x = self.encoder(x, training=training)
        x = self.decoder(x, training=training)
        if self.return_heatmap:
            return x
        else:
            return x


def get_lwopenpose(calc_3d_features,
                   keypoints,
                   model_name=None,
                   pretrained=False,
                   root=os.path.join("~", ".tensorflow", "models"),
                   **kwargs):
    """
    Create Lightweight OpenPose 2D/3D model with specific parameters.

    Parameters:
    ----------
    calc_3d_features : bool, default False
        Whether to calculate 3D features.
    keypoints : int
        Number of keypoints.
    model_name : str or None, default None
        Model name for loading pretrained model.
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    root : str, default '~/.tensorflow/models'
        Location for keeping the model parameters.
    """
    encoder_channels = [[64], [128, 128], [256, 256, 512, 512, 512, 512, 512, 512]]
    encoder_paddings = [[1], [1, 1], [1, 1, 1, 2, 1, 1, 1, 1]]
    encoder_init_block_channels = 32
    encoder_final_block_channels = 128
    refinement_units = 1

    net = LwOpenPose(
        encoder_channels=encoder_channels,
        encoder_paddings=encoder_paddings,
        encoder_init_block_channels=encoder_init_block_channels,
        encoder_final_block_channels=encoder_final_block_channels,
        refinement_units=refinement_units,
        calc_3d_features=calc_3d_features,
        keypoints=keypoints,
        **kwargs)

    if pretrained:
        if (model_name is None) or (not model_name):
            raise ValueError("Parameter `model_name` should be properly initialized for loading pretrained model.")
        from .model_store import get_model_file
        in_channels = kwargs["in_channels"] if ("in_channels" in kwargs) else 3
        input_shape = (1,) + (in_channels,) + net.in_size if net.data_format == "channels_first" else\
            (1,) + net.in_size + (in_channels,)
        net.build(input_shape=input_shape)
        net.load_weights(
            filepath=get_model_file(
                model_name=model_name,
                local_model_store_dir_path=root))

    return net


def lwopenpose2d_mobilenet_cmupan_coco(keypoints=19, data_format="channels_last", **kwargs):
    """
    Lightweight OpenPose 2D model on the base of MobileNet for CMU Panoptic from 'Real-time 2D Multi-Person Pose
    Estimation on CPU: Lightweight OpenPose,' https://arxiv.org/abs/1811.12004.

    Parameters:
    ----------
    keypoints : int, default 19
        Number of keypoints.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    root : str, default '~/.tensorflow/models'
        Location for keeping the model parameters.
    """
    return get_lwopenpose(calc_3d_features=False, keypoints=keypoints, model_name="lwopenpose2d_mobilenet_cmupan_coco",
                          data_format=data_format, **kwargs)


def lwopenpose3d_mobilenet_cmupan_coco(keypoints=19, data_format="channels_last", **kwargs):
    """
    Lightweight OpenPose 3D model on the base of MobileNet for CMU Panoptic from 'Real-time 2D Multi-Person Pose
    Estimation on CPU: Lightweight OpenPose,' https://arxiv.org/abs/1811.12004.

    Parameters:
    ----------
    keypoints : int, default 19
        Number of keypoints.
    data_format : str, default 'channels_last'
        The ordering of the dimensions in tensors.
    pretrained : bool, default False
        Whether to load the pretrained weights for model.
    root : str, default '~/.tensorflow/models'
        Location for keeping the model parameters.
    """
    return get_lwopenpose(calc_3d_features=True, keypoints=keypoints, model_name="lwopenpose3d_mobilenet_cmupan_coco",
                          data_format=data_format, **kwargs)


def _test():
    import numpy as np
    import tensorflow.keras.backend as K

    # os.environ["TF_CUDNN_DETERMINISTIC"] = "1"
    # os.environ["TF_DETERMINISTIC_OPS"] = "1"

    data_format = "channels_last"
    # data_format = "channels_first"
    in_size_ = (368, 368)
    keypoints = 19
    return_heatmap = True
    pretrained = False

    models = [
        (lwopenpose2d_mobilenet_cmupan_coco, "2d", in_size_),
        (lwopenpose3d_mobilenet_cmupan_coco, "3d", in_size_),
    ]

    for model, model_dim, in_size in models:

        net = model(pretrained=pretrained, in_size=in_size, return_heatmap=return_heatmap, data_format=data_format)

        batch = 14
        x = tf.random.normal((batch, 3, in_size[0], in_size[1]) if is_channels_first(data_format) else
                             (batch, in_size[0], in_size[1], 3))
        y = net(x)
        assert (y.shape[0] == batch)
        keypoints_ = 3 * keypoints if model_dim == "2d" else 6 * keypoints
        if is_channels_first(data_format):
            assert ((y.shape[1] == keypoints_) and (y.shape[2] == x.shape[2] // 8) and
                    (y.shape[3] == x.shape[3] // 8))
        else:
            assert ((y.shape[3] == keypoints_) and (y.shape[1] == x.shape[1] // 8) and
                    (y.shape[2] == x.shape[2] // 8))

        weight_count = sum([np.prod(K.get_value(w).shape) for w in net.trainable_weights])
        print("m={}, {}".format(model.__name__, weight_count))
        assert (model != lwopenpose2d_mobilenet_cmupan_coco or weight_count == 4091698)
        assert (model != lwopenpose3d_mobilenet_cmupan_coco or weight_count == 5085983)


if __name__ == "__main__":
    _test()
